"""
iniparser parses ini files into a python dictionary of dictionaries. Duplicate
keys in a section convert the value to a list. Numbers are automatically
converted to python ints or floats. Line continuations based on hanging indents
are supported. Sections inherit keys from the `[DEFAULT]` section. All keys are
converted to lower case. Variable interpolation is not supported.
"""
import string
from parsr import (Char, EOF, EOL, LeftBracket, Many, Number, OneLineComment,
                   Opt, Parser, RightBracket, String, WithIndent, WS)


class HangingString(Parser):
    def __init__(self, chars):
        super().__init__()
        self.add_child(String(chars) << (EOL | EOF))

    def process(self, pos, data, ctx):
        old = pos
        results = []
        while True:
            try:
                if ctx.col(pos) > ctx.indents[-1]:
                    pos, res = self.children[0].process(pos, data, ctx)
                    results.append(res)
                else:
                    pos = old
                    break
                old = pos
                pos, _ = WS.process(pos, data, ctx)
            except Exception:
                break
        ret = " ".join(results)
        return pos, ret


def to_dict(x):
    x = dict([i for i in x if i is not None])
    d = {}
    for k, v in x.items():
        k = k.lower()
        if k in d:
            if not isinstance(d[k], list):
                d[k] = [d[k]]
            d[k].append(v)
        else:
            d[k] = v
    return d


header_chars = (set(string.printable) - set(string.whitespace) - set("[]"))
key_chars = header_chars - set("=:")
value_chars = set(string.printable) - set("\n\r")

LineEnd = EOL | EOF
LeftEnd = (WS + LeftBracket + WS)
RightEnd = (WS + RightBracket + WS)
Header = LeftEnd >> String(header_chars) << RightEnd
Key = WS >> String(key_chars) << WS
Sep = Char("=") | Char(":")
Value = WS >> ((Number << LineEnd) | HangingString(value_chars))
KVPair = WithIndent(Key + Opt(Sep + Value, default=[None, None])).map(lambda a: (a[0], a[1][1]))
Comment = (WS >> (OneLineComment("#") | OneLineComment(";")).map(lambda x: None))
Line = Comment | KVPair
Section = Header + (Many(Line).map(to_dict))
Doc = Many(Comment | Section).map(to_dict)
Top = Doc + EOF


def loads(s):
    res = Top(s)[0]
    default = res.get("default")
    if default:
        default_keys = set(default)
        for header, items in res.items():
            if header != "default":
                missing = default_keys - set(items)
                for m in missing:
                    items[m] = default[m]

    return res


def load(f):
    return loads(f.read())



"""
json_parser handles primitive json parsing. It doesn't handle unicode or
numbers in scientific notation.
"""
from parsr import (Colon, Comma, EOF, Forward, Literal, LeftBracket, LeftCurly,
                   Number, RightBracket, RightCurly, QuotedString, WS)


JsonArray = Forward()
JsonObject = Forward()
TRUE = Literal("true", value=True) % "true"
FALSE = Literal("false", value=False) % "false"
NULL = Literal("null", value=None) % "null"
SimpleValue = (Number | QuotedString | JsonObject | JsonArray | TRUE | FALSE | NULL)
JsonValue = (WS >> SimpleValue << WS)
Key = (QuotedString << Colon)
KVPairs = (((WS >> Key) + JsonValue).sep_by(Comma))
JsonArray <= (LeftBracket >> JsonValue.sep_by(Comma) << RightBracket)
JsonObject <= (LeftCurly >> KVPairs.map(lambda res: {k: v for (k, v) in res}) << RightCurly)
Top = JsonValue + EOF


def loads(data):
    return Top(data)[0]


def load(f):
    return loads(f.read())



from parsr.httpd_conf import loads


DATA = r"""
<IfModule log_config_module>
    #
    # The following directives define some format nicknames for use with
    # a CustomLog directive (see below).
    #
    LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\"" combined
    LogFormat "%h %l %u %t \"%r\" %>s %b" common
</IfModule>
    <IfModule logio_module>
      # You need to enable mod_logio.c to use %I and %O
      LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" %I %O" combinedio
    </IfModule>

    #
    # The location and format of the access logfile (Common Logfile Format).
    # If you do not define any access logfiles within a <VirtualHost>
    # container, they will be logged here.  Contrariwise, if you *do*
    # define per-<VirtualHost> access logfiles, transactions will be
    # logged therein and *not* in this file.
    #
    #CustomLog "logs/access_log" common
    CustomLog logs/ssl_request_log \
    "%t %h %{SSL_PROTOCOL}x %{SSL_CIPHER}x "%r" %b"

    #
    # If you prefer a logfile with access, agent, and referer information
    # (Combined Logfile Format) you can use the following directive.
    #
    CustomLog "logs/access_log" combined
"""


def test_if_module():
    val = loads(DATA)
    assert len(val) == 4



from parsr.iniparser import loads


DATA = """
[DEFAULT]
trooper_guns = miss

[global]
logging=debug
log=/var/logs/sample.log

# Keep this info secret
[secret_stuff]
username=dvader
password=luke_is_my_son

[facts]
major_vulnerability=ray-shielded particle exhaust vent
vader = definitely Luke's
    father

[settings]
music=The Imperial March
color=black
"""


def test_iniparser():
    res = loads(DATA)
    assert len(res) == 5


def test_hanging_indent():
    res = loads(DATA)
    assert res["facts"]["vader"] == "definitely Luke's father"


def test_defaults():
    res = loads(DATA)
    assert res["facts"]["trooper_guns"] == "miss"



"""
corosync_conf parses corosync configuration files into nested dictionaries.
"""
import string
from parsr import (Char, EOF, EOL, Forward, LeftCurly, Literal, RightCurly,
                   Many, Number, OneLineComment, String, QuotedString, WS)


def skip_none(x):
    return [i for i in x if i is not None]


Stmt = Forward()
LineEnd = (EOL | EOF) % "LineEnd"
Num = (Number << LineEnd) % "Num"
NULL = Literal("none", value=None) % "none"
Sep = (Char(":") | Char("=")) % "Sep"
Comment = (WS >> OneLineComment("#").map(lambda x: None)) % "Comment"
BeginBlock = (WS >> LeftCurly << WS)
EndBlock = (WS >> RightCurly << WS)
Bare = String(set(string.printable) - (set(string.whitespace) | set("#{}'\"")))
Name = WS >> String(string.ascii_letters + "_") << WS
Value = WS >> (Num | NULL | QuotedString | Bare) << WS
Block = BeginBlock >> Many(Stmt).map(skip_none).map(dict) << EndBlock
Stanza = (Name + (Block | (Sep >> Value))) | Comment
Stmt <= WS >> Stanza << WS
Doc = Many(Stmt).map(skip_none).map(dict)
Top = Doc + EOF


def loads(data):
    return Top(data)[0]


def load(f):
    return loads(f.read())



import string
from parsr import (Char, EOF, EOL, EndTagName, Forward, FS, GT, LT, Letters,
                   Many, Number, OneLineComment, QuotedString, StartTagName,
                   String, WS, WSChar)


def skip_none(x):
    return [i for i in x if i is not None]


Complex = Forward()
Cont = Char("\\") + EOL
StartName = WS >> StartTagName(Letters) << WS
EndName = WS >> EndTagName(Letters) << WS
Comment = (WS >> OneLineComment("#")).map(lambda x: None)
AttrStart = Many(WSChar)
AttrEnd = (Many(WSChar) + Cont) | Many(WSChar)
BareAttr = String(set(string.printable) - (set(string.whitespace) | set("#;{}<>\\'\"")))
Attr = AttrStart >> (Number | BareAttr | QuotedString) << AttrEnd
Attrs = Many(Attr)
StartTag = (WS + LT) >> (StartName + Attrs) << (GT + WS)
EndTag = (WS + LT + FS) >> EndName << (GT + WS)
Simple = WS >> (Letters + Attrs) << WS
Stanza = Simple | Complex | Comment
Complex <= StartTag + Many(Stanza).map(skip_none) << EndTag
Doc = Many(Stanza).map(skip_none)
Top = Doc + EOF


def loads(data):
    return Top(data)[0]


def load(f):
    return loads(f.read())



import string
from parsr import (EOF, Forward, LeftCurly, Lift, RightCurly, Many, Number,
                   OneLineComment, SemiColon, SingleQuotedString, String, WS)


class Value:
    def __init__(self, name, attrs, body):
        self.name = name
        self.attrs = attrs
        self.body = body if body != ";" else []


def skip_none(x):
    return [i for i in x if i is not None]


Stmt = Forward()
Comment = OneLineComment("#").map(lambda x: None)
BeginBlock = WS >> LeftCurly << WS
EndBlock = WS >> RightCurly << WS
Bare = String(set(string.printable) - (set(string.whitespace) | set("#;{}'\"")))
Name = WS >> String(string.ascii_letters + "_") << WS
Attr = WS >> (Number | Bare | SingleQuotedString) << WS
Attrs = Many(Attr)
Block = BeginBlock >> Many(Stmt).map(skip_none) << EndBlock
Stanza = (Lift(Value) * Name * Attrs * (Block | SemiColon)) | Comment
Stmt <= WS >> Stanza << WS
Doc = Many(Stmt).map(skip_none)
Top = Doc + EOF


def loads(data):
    return Top(data)[0]


def load(f):
    return loads(f.read())
